#define _USE_MATH_DEFINES
#include <cmath>
#include "ambientocclusion.h"
#include "blend.h"
#include "shader.h"
#include "tests.h"

namespace {

const char occlude_fs[] =
	"uniform sampler2D depth;\n"
	"uniform sampler2D rotate;\n"
	"uniform vec2 screen_size;\n"
	"uniform vec2 depth_ratio;\n"
	"uniform float darkness;\n"
	"varying vec2 texcoord;\n"
	"void main()\n"
	"{\n"
	"	mat2 transform = mat2(texture2D(rotate, texcoord*screen_size/4.0)*2.0-1.0)\n"
 	"		*mat2(screen_size.y/screen_size.x, 0.0, 0.0, 1.0)*3.0/screen_size.y;\n"
	"	float sample = depth_ratio.x/(texture2D(depth, texcoord).r-depth_ratio.y);\n"
	"	float sum = 0.0;\n"
	"	for(int i=0; i<=3; ++i)\n"
	"		for(int j=0; j<=3; ++j)\n"
	"		{\n"
	"			vec2 offs = transform*vec2(float(i)-1.5, float(j)-1.5);\n"
	"			float dxy = length(offs)*-sample;\n"
	"			float dz = depth_ratio.x/(texture2D(depth, texcoord+offs).r-depth_ratio.y)-sample;\n"
	"			if(abs(dz)<3.0*dxy)\n"
	"				sum += atan(dz/dxy)/1.570796;\n"
	"			else if(dz<0.0)\n"
	"				sum -= 0.8;\n"
	"		}\n"
	"	gl_FragColor = vec4(min(1.0-sum*darkness/16.0, 1.0), 0.0, 0.0, 1.0);\n"
	"}\n";

const char combine_fs[] =
	"uniform sampler2D color;\n"
	"uniform sampler2D depth;\n"
	"uniform sampler2D occlusion;\n"
	"uniform vec2 screen_size;\n"
	"uniform vec2 depth_ratio;\n"
	"varying vec2 texcoord;\n"
	"void main()\n"
	"{\n"
	"	float sample = depth_ratio.x/(texture2D(depth, texcoord).r-depth_ratio.y);\n"
	"	float sum = 1.0;\n"
	"	float count = 1.0;\n"
	"	for(int i=0; i<=3; ++i)\n"
	"		for(int j=0; j<=3; ++j)\n"
	"		{\n"
	"			vec2 offs = vec2(float(i)-1.5, float(j)-1.5)/screen_size;\n"
	"			float dxy = length(offs)*-sample;\n"
	"			float dz = depth_ratio.x/(texture2D(depth, texcoord+offs).r-depth_ratio.y)-sample;\n"
	"			if(abs(dz)<3.0*dxy)\n"
	"			{\n"
	"				sum += texture2D(occlusion, texcoord+offs).r;\n"
	"				count += 1.0;\n"
	"			}\n"
	"		}\n"
	"	gl_FragColor = texture2D(color, texcoord)*sum/count;\n"
	"}\n";

}

namespace Msp {
namespace GL {

AmbientOcclusion::AmbientOcclusion(unsigned w, unsigned h, float depth_ratio):
	quad(get_fullscreen_quad())
{
	occlude_shader.attach_shader(get_fullscreen_vertex_shader());
	occlude_shader.attach_shader_owned(new FragmentShader(occlude_fs));
	occlude_shader.bind_attribute(get_component_type(VERTEX2), "vertex");
	occlude_shader.link();

	combine_shader.attach_shader(get_fullscreen_vertex_shader());
	combine_shader.attach_shader_owned(new FragmentShader(combine_fs));
	combine_shader.bind_attribute(get_component_type(VERTEX2), "vertex");
	combine_shader.link();

	occlusion.storage(RGB, w, h);
	occlusion.set_min_filter(NEAREST);
	occlusion.set_mag_filter(NEAREST);
	occlusion.set_wrap(CLAMP_TO_EDGE);
	fbo.attach(COLOR_ATTACHMENT0, occlusion, 0);
	fbo.require_complete();

	combine_texturing.attach(2, occlusion);

	rotate_lookup.storage(RGBA, 4, 4);
	rotate_lookup.set_min_filter(NEAREST);
	rotate_lookup.set_mag_filter(NEAREST);
	unsigned char data[64];
	for(unsigned i=0; i<16; ++i)
	{
		float a = ((i*541)%16)*M_PI/32;
		float c = cos(a);
		float s = sin(a);
		data[i*3  ] = static_cast<unsigned char>(127+c*127);
		data[i*3+1] = static_cast<unsigned char>(127+s*127);
		data[i*3+2] = static_cast<unsigned char>(127-s*127);
		data[i*3+4] = static_cast<unsigned char>(127+c*127);
	}
	rotate_lookup.image(0, RGBA, UNSIGNED_BYTE, data);

	occlude_texturing.attach(1, rotate_lookup);

	occlude_shdata.uniform("depth", 0);
	occlude_shdata.uniform("rotate", 1);
	occlude_shdata.uniform("screen_size", static_cast<float>(w), static_cast<float>(h));

	combine_shdata.uniform("color", 1);
	combine_shdata.uniform("depth", 0);
	combine_shdata.uniform("occlusion", 2);
	combine_shdata.uniform("screen_size", static_cast<float>(w), static_cast<float>(h));

	set_depth_ratio(depth_ratio);
	set_darkness(1.5);
}

void AmbientOcclusion::set_depth_ratio(float depth_ratio)
{
	depth_ratio = 1/depth_ratio;

	occlude_shdata.uniform("depth_ratio", depth_ratio, 1+depth_ratio);
	combine_shdata.uniform("depth_ratio", depth_ratio, 1+depth_ratio);
}

void AmbientOcclusion::set_darkness(float darkness)
{
	occlude_shdata.uniform("darkness", darkness);
}

void AmbientOcclusion::render(const Texture2D &color, const Texture2D &depth)
{
	occlude_texturing.attach(0, depth);
	combine_texturing.attach(0, depth);
	combine_texturing.attach(1, color);

	BindRestore unbind_dtest(static_cast<DepthTest *>(0));
	BindRestore unbind_blend(static_cast<Blend *>(0));
	Bind bind_mesh(quad);

	{
		BindRestore bind_fbo(fbo);
		Bind bind_tex(occlude_texturing);
		Bind bind_shader(occlude_shader);
		occlude_shdata.apply();
		quad.draw();
	}

	Bind bind_tex(combine_texturing);
	Bind bind_shader(combine_shader);
	combine_shdata.apply();
	quad.draw();
}

} // namespace GL
} // namespace Msp
