#include <cmath>
#include "environmentmap.h"
#include "renderer.h"
#include "texunit.h"

namespace Msp {
namespace GL {

EnvironmentMap::EnvironmentMap(unsigned s, Renderable &r, Renderable &e):
	Effect(r),
	size(s),
	environment(e),
	rendered(false)
{
	env_tex.storage(RGB, size);
	env_tex.set_wrap(CLAMP_TO_EDGE);
	env_tex.set_min_filter(LINEAR);
	depth_buf.storage(DEPTH_COMPONENT, size, size);
	for(unsigned i=0; i<6; ++i)
	{
		fbo[i].attach(COLOR_ATTACHMENT0, env_tex, TextureCube::enumerate_faces(i), 0);
		fbo[i].attach(DEPTH_ATTACHMENT, depth_buf);
		fbo[i].require_complete();
	}

	// XXX Make the depth range configurable
	camera.set_field_of_view(Geometry::Angle<float>::right());
	camera.set_aspect(1);
	camera.set_depth_clip(0.1, 100);

}

void EnvironmentMap::setup_frame() const
{
	if(rendered)
		return;

	rendered = true;
	renderable.setup_frame();
	environment.setup_frame();

	Renderer env_renderer(0);
	const Matrix *matrix = renderable.get_matrix();
	Vector3 position = (*matrix)*Vector3();
	camera.set_position(position);
	for(unsigned i=0; i<6; ++i)
	{
		TextureCubeFace face = TextureCube::enumerate_faces(i);
		fbo[i].bind();
		fbo[i].clear(COLOR_BUFFER_BIT|DEPTH_BUFFER_BIT);
		camera.set_look_direction(env_tex.get_face_direction(face));
		camera.set_up_direction(env_tex.get_t_direction(face));
		env_renderer.begin(&camera);
		env_renderer.exclude(renderable);
		env_renderer.exclude(*this);
		env_renderer.render(environment);
	}
	Framebuffer::unbind();
}

void EnvironmentMap::finish_frame() const
{
	if(rendered)
	{
		rendered = false;
		renderable.finish_frame();
		environment.finish_frame();
	}
}

void EnvironmentMap::render(Renderer &renderer, const Tag &tag) const
{
	if(!enabled_passes.count(tag))
		return renderer.render(renderable, tag);

	Renderer::Push _push_rend(renderer);

	unsigned unit = renderer.allocate_effect_texunit();
	shdata.uniform("environment", static_cast<int>(unit));
	Bind _bind_env(env_tex, unit);

	const Matrix &camera_matrix = renderer.get_camera()->get_object_matrix();
	shdata.uniform("env_eye_matrix", camera_matrix.block<3, 3>(0, 0));

	renderer.add_shader_data(shdata);
	renderer.render(renderable, tag);
}

} // namespace GL
} // namespace Msp
